import abc
from typing import Type

from pydantic import BaseModel


class CryptoParser(BaseModel):
    __metaclass__ = abc.ABCMeta

    @classmethod
    @abc.abstractmethod
    def from_cryptography(
        cls: Type["CryptoParser"], crypto_obj
    ) -> "CryptoParser":
        """
        Parses a cryptography x509 object into a CryptoParser

        Arguments:
             crypto_obj --The cryptography object
        """



import base64

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives._serialization import (
    Encoding,
    PublicFormat,
)
from cryptography.hazmat.primitives.hashes import (
    SHA256,
    SHA1,
    SHA512,
    SHA384,
    SHA224,
)
from cryptography.x509 import ocsp
from loguru import logger

from pki_tools.types.chain import Chain
from pki_tools.types.certificate import Certificate
from pki_tools.utils import HTTPX_CLIENT, verify_signature
from pki_tools.exceptions import (
    ExtensionMissing,
    OcspInvalidResponseStatus,
    OcspFetchFailure,
    Error,
)
from pki_tools.types.extensions import AuthorityInformationAccess
from pki_tools.types.ocsp import OCSPResponse

OCSP_ALGORITHMS_TO_CHECK = [SHA256(), SHA1(), SHA512(), SHA224(), SHA384()]


def _is_revoked_multiple_issuers(
    cert: Certificate,
    cert_issuer: Chain,
    ocsp_issuer: Chain,
):
    cert_issuer.check_chain()
    ocsp_issuer.check_chain()

    issuer = cert_issuer.get_issuer(cert)

    log = logger.bind(
        cert=cert.pem_string,
        serial=cert.serial_number,
    )

    if cert.extensions.authority_information_access is None:
        log.debug("OCSP extension missing")
        raise ExtensionMissing()

    for i, alg in enumerate(OCSP_ALGORITHMS_TO_CHECK):
        try:
            req_path = _construct_req_path(cert, issuer, alg)

            return _check_ocsp_status(
                cert.extensions.authority_information_access,
                req_path,
                cert,
                ocsp_issuer,
            )
        except OcspInvalidResponseStatus:
            log.bind(alg=alg.name).debug(
                "OCSP check failed, trying another algorithm"
            )
            if i + 1 == len(OCSP_ALGORITHMS_TO_CHECK):
                log.bind(
                    algs=[alg.name for alg in OCSP_ALGORITHMS_TO_CHECK]
                ).debug("All algorithms check failed")
                raise

    return False


def _construct_req_path(cert, issuer_cert, alg):
    cert = cert._x509_obj
    issuer_cert = issuer_cert._x509_obj
    builder = ocsp.OCSPRequestBuilder()
    builder = builder.add_certificate(cert, issuer_cert, alg)
    req = builder.build()
    return base64.b64encode(
        req.public_bytes(serialization.Encoding.DER)
    ).decode()


def _check_ocsp_status(
    aia: AuthorityInformationAccess,
    req_path,
    cert: Certificate,
    issuer_chain: Chain,
):
    log = logger.bind(serial=cert.hex_serial)

    checked_status = False
    for access_description in aia:
        if access_description.access_method != "OCSP":
            continue

        checked_status = True

        server = access_description.access_location

        ocsp_res = _get_ocsp_status(f"{server}/{req_path}")

        _verify_ocsp_signature(ocsp_res, issuer_chain)

        if ocsp_res.is_revoked:
            log.bind(
                date=str(ocsp_res._x509_obj.revocation_time),
            ).debug("Certificate revoked")
            return True

    if not checked_status:
        raise ExtensionMissing()

    log.debug("Certificate valid")
    return False


def _get_ocsp_status(uri) -> OCSPResponse:
    ret = HTTPX_CLIENT.get(
        uri, headers={"Content-Type": "application/ocsp-request"}
    )

    log = logger.bind(status=ret.status_code)
    if ret.status_code != 200:
        log.error("OCSP status fetch failed")
        raise OcspFetchFailure(
            f"Unexpected response status code: {ret.status_code}"
        )

    ocsp_res = ocsp.load_der_ocsp_response(ret.content)
    ocsp_res = OCSPResponse.from_cryptography(ocsp_res)

    if not ocsp_res.is_successful:
        log.bind(res=ocsp_res.response_status).debug("Invalid OCSP response")
        raise OcspInvalidResponseStatus(
            f"Invalid OCSP Response status: {ocsp_res.response_status}"
        )

    return ocsp_res


def _verify_ocsp_signature(ocsp_response: OCSPResponse, issuer_chain: Chain):
    for issuer_cert in issuer_chain.certificates:
        der_key = issuer_cert.public_key.public_bytes(
            encoding=Encoding.DER,
            format=PublicFormat.PKCS1,
        )
        cert_public_hash = ocsp_response.hash_with_alg(der_key)

        if cert_public_hash == ocsp_response.issuer_key_hash:
            break
    else:
        logger.error("Couldn't find OCSP response issuer")
        raise Error("Couldn't find OCSP response issuer")

    verify_signature(ocsp_response, issuer_cert)



import typing
from typing import List, Optional, Iterable, Union

from cryptography import x509
from cryptography.hazmat._oid import ExtensionOID
from cryptography.hazmat.bindings._rust import ObjectIdentifier


from cryptography.x509.extensions import (
    ExtensionNotFound,
    ExtensionTypeVar,
)

from loguru import logger
from pydantic import BaseModel, Field, ConfigDict

from pki_tools.types.name import Name
from pki_tools.types.utils import _byte_to_hex


class Extension(BaseModel):
    critical: Optional[bool] = False

    @property
    def name(self):
        name = "".join(
            [" " + c if c.isupper() else c for c in self.__class__.__name__]
        )[1:]
        if self.critical:
            name += " (critical)"
        return name


class AuthorityKeyIdentifier(Extension):
    key_identifier: Optional[bytes]
    authority_cert_issuer: Optional[List[str]]
    authority_cert_serial_number: Optional[int]

    @classmethod
    def from_cryptography(cls, extension: x509.AuthorityKeyIdentifier):
        issuers = []
        if extension.authority_cert_issuer is not None:
            for general_name in extension.authority_cert_issuer:
                issuers.append(general_name.value)

        opt_issuers = None
        if issuers:
            opt_issuers = issuers
        return cls(
            key_identifier=extension.key_identifier,
            authority_cert_issuer=opt_issuers,
            authority_cert_serial_number=extension.authority_cert_serial_number,
        )

    def _string_dict(self):
        ret = {}

        if self.key_identifier is not None:
            hex_key = _byte_to_hex(self.key_identifier)
            ret["Key Identifier"] = hex_key
        if self.authority_cert_issuer is not None:
            ret["Authority Cert Issuer"] = self.authority_cert_issuer
        if self.authority_cert_serial_number is not None:
            ret[
                "Authority Cert Serial Number"
            ] = self.authority_cert_serial_number

        if ret:
            return {self.name: ret}
        return {}


class SubjectKeyIdentifier(Extension):
    subject_key_identifier: bytes

    @classmethod
    def from_cryptography(cls, extension: x509.SubjectKeyIdentifier):
        return cls(subject_key_identifier=extension.key_identifier)

    def _string_dict(self):
        hex_key = _byte_to_hex(self.subject_key_identifier)
        return {self.name: {"Subject Key Identifier": hex_key}}


class KeyUsage(Extension):
    digital_signature: bool = False
    content_commitment: bool = False
    key_encipherment: bool = False
    data_encipherment: bool = False
    key_agreement: bool = False
    key_cert_sign: bool = False
    crl_sign: bool = False
    encipher_only: bool = False
    decipher_only: bool = False

    @classmethod
    def from_cryptography(cls, extension: x509.KeyUsage):
        try:
            encipher_only = extension.encipher_only
        except ValueError:
            encipher_only = False
        try:
            decipher_only = extension.decipher_only
        except ValueError:
            decipher_only = False

        return cls(
            digital_signature=extension.digital_signature,
            content_commitment=extension.content_commitment,
            key_encipherment=extension.key_encipherment,
            data_encipherment=extension.data_encipherment,
            key_agreement=extension.key_agreement,
            key_cert_sign=extension.key_cert_sign,
            crl_sign=extension.crl_sign,
            encipher_only=encipher_only,
            decipher_only=decipher_only,
        )

    def _string_dict(self):
        true_fields = []
        for field in self.model_fields:
            if field == "critical":
                continue

            field_title = " ".join(ele.title() for ele in field.split("_"))
            if getattr(self, field):
                true_fields.append(field_title)

        return {self.name: ", ".join(true_fields)}


class NoticeReference(Extension):
    organization: str
    notice_numbers: List[int]

    @classmethod
    def from_cryptography(cls, notice_reference: x509.NoticeReference):
        return cls(
            organization=notice_reference.organization,
            notice_numbers=notice_reference.notice_numbers,
        )

    def _string_dict(self):
        return {
            self.name: {
                "Organization": self.organization,
                "Notice Numbers": self.notice_numbers,
            }
        }


class UserNotice(Extension):
    notice_reference: Optional[NoticeReference]
    explicit_text: Optional[str]

    @classmethod
    def from_cryptography(cls, policy_info: x509.UserNotice):
        return cls(
            notice_reference=NoticeReference.from_cryptography(
                policy_info.notice_reference
            ),
            explicit_text=policy_info.explicit_text,
        )

    def _string_dict(self):
        ret = {self.name: {}}

        if self.notice_reference is not None:
            ret[self.name].update(self.notice_reference._string_dict())
        if self.explicit_text is not None:
            ret[self.name]["Explicit Text"] = self.explicit_text
        return ret


class PolicyInformation(Extension):
    policy_identifier: str
    policy_qualifiers: Optional[List[Union[str, UserNotice]]]

    @classmethod
    def from_cryptography(cls, policy_info: x509.PolicyInformation):
        policy_qualifiers = None
        if policy_info.policy_qualifiers is not None:
            policy_qualifiers = []
            for qualifier in policy_info.policy_qualifiers:
                if isinstance(qualifier, str):
                    policy_qualifiers.append(f"CPS: {qualifier}")
                else:
                    policy_qualifiers.append(
                        UserNotice.from_cryptography(qualifier)
                    )

        return cls(
            policy_identifier=policy_info.policy_identifier.dotted_string,
            policy_qualifiers=policy_qualifiers,
        )

    def _string_dict(self):
        name = f"Policy {self.policy_identifier}"
        ret = {name: []}

        if self.policy_qualifiers is not None:
            for qualifier in self.policy_qualifiers:
                if isinstance(qualifier, str):
                    ret[name].append(qualifier)
                else:
                    ret[name].append(qualifier._string_dict())
        return ret


class CertificatePolicies(Extension):
    policy_information: List[PolicyInformation]

    @classmethod
    def from_cryptography(cls, extension: x509.CertificatePolicies):
        res = []
        for policy_information in extension:
            info = PolicyInformation.from_cryptography(policy_information)
            res.append(info)

        return cls(policy_information=res)

    def _string_dict(self):
        ret = {self.name: []}
        for policy_information in self.policy_information:
            ret[self.name].append(policy_information._string_dict())

        return ret


class AlternativeName(Extension):
    general_names: List[str]

    @classmethod
    def from_cryptography(cls, extension: x509.SubjectAlternativeName):
        names = []
        for general_name in extension:
            names.append(_general_name_to_str(general_name))

        return cls(general_names=names)

    def _string_dict(self):
        return {self.name: self.general_names}


class SubjectAlternativeName(AlternativeName):
    pass


class IssuerAlternativeName(AlternativeName):
    pass


class SubjectDirectoryAttributes(Extension):
    attributes: List[str]

    @classmethod
    def from_cryptography(cls, extension: x509.UnrecognizedExtension):
        attributes = []
        vals = extension.value
        if not isinstance(vals, List):
            vals = [vals]

        for val in vals:
            if isinstance(val, bytes):
                val = _byte_to_hex(val)

            attributes.append(val)

        return cls(attributes=attributes)

    def _string_dict(self):
        return {self.name: self.attributes}


class BasicConstraints(Extension):
    ca: bool
    path_len_constraint: Optional[int]

    @classmethod
    def from_cryptography(cls, extension: x509.BasicConstraints):
        return cls(ca=extension.ca, path_len_constraint=extension.path_length)

    def _string_dict(self):
        ret = {
            self.name: {
                "CA": self.ca,
            }
        }

        if self.path_len_constraint is not None:
            ret["Path Lenght"] = self.path_len_constraint

        return ret


class NameConstraints(Extension):
    permitted_subtrees: Optional[List[str]]
    excluded_subtrees: Optional[List[str]]

    @classmethod
    def from_cryptography(cls, extension: x509.NameConstraints):
        permitted_subtrees = []
        for permitted in extension.permitted_subtrees:
            permitted_subtrees.append(_general_name_to_str(permitted))
        if not permitted_subtrees:
            permitted_subtrees = None

        excluded_subtrees = []
        for excluded in extension.excluded_subtrees:
            excluded_subtrees.append(_general_name_to_str(excluded))
        if not excluded_subtrees:
            excluded_subtrees = None

        return cls(
            permitted_subtrees=permitted_subtrees,
            excluded_subtrees=excluded_subtrees,
        )

    def _string_dict(self):
        return {
            self.name: {
                "Permitted Subtrees": self.permitted_subtrees,
                "Excluded Subtrees": self.excluded_subtrees,
            }
        }


class PolicyConstraints(Extension):
    require_explicit_policy: Optional[int]
    inhibit_policy_mapping: Optional[int]

    @classmethod
    def from_cryptography(cls, extension: x509.PolicyConstraints):
        return cls(
            require_explicit_policy=extension.require_explicit_policy,
            inhibit_policy_mapping=extension.inhibit_policy_mapping,
        )

    def _string_dict(self):
        ret = {self.name: {}}

        if self.require_explicit_policy is not None:
            ret[self.name][
                "Require Explicit Policy"
            ] = self.require_explicit_policy
        if self.inhibit_policy_mapping is not None:
            ret[self.name][
                "Inhibit Policy Mapping"
            ] = self.inhibit_policy_mapping

        return ret


EKU_OID_MAPPING = {
    "1.3.6.1.5.5.7.3.1": "Server Authentication",
    "1.3.6.1.5.5.7.3.2": "Client Authentication",
    "1.3.6.1.5.5.7.3.3": "Code Signing",
    "1.3.6.1.5.5.7.3.4": "Email Protection",
    "1.3.6.1.5.5.7.3.8": "Time Stamping",
    "1.3.6.1.5.5.7.3.9": "OCSP Signing",
    "1.3.6.1.5.5.7.3.17": "IPsec IKE",
    "2.5.29.37.0": "Any Extended Key Usage",
    "1.3.6.1.4.1.311.20.2.2": "Smart Card Logon",
    "1.3.6.1.5.2.3.5": "Kerberos PKINIT KDC",
    "1.3.6.1.4.1.11129.2.4.4": "Certificate Transparency",
}


class ExtendedKeyUsage(Extension):
    ext_key_usage_syntax: List[str]

    @classmethod
    def from_cryptography(cls, extension: x509.ExtendedKeyUsage):
        ext_key_usage_syntax = []
        for key_usage in extension:
            name = EKU_OID_MAPPING.get(
                key_usage.dotted_string,
                f"Unknown OID ({key_usage.dotted_string})",
            )
            ext_key_usage_syntax.append(name)
        return cls(ext_key_usage_syntax=ext_key_usage_syntax)

    def _string_dict(self):
        return {self.name: self.ext_key_usage_syntax}


class DistributionPoint(Extension):
    full_name: Optional[List[str]]
    name_relative_to_crl_issuer: Optional[List[str]]
    reasons: Optional[List[str]]
    crl_issuer: Optional[List[str]]

    @classmethod
    def from_cryptography(cls, extension: x509.DistributionPoint):
        full_names = None
        if extension.full_name is not None:
            full_names = []
            for full_name in extension.full_name:
                full_names.append(_general_name_to_str(full_name))

        relative_names = None
        if extension.relative_name is not None:
            relative_names = []
            for rel_name in extension.relative_name:
                relative_names.append(rel_name.rfc4514_string())

        reasons = None
        if extension.reasons is not None:
            reasons = []
            for reason in extension.reasons:
                reasons.append(reason.value)

        crl_issuers = None
        if extension.crl_issuer is not None:
            crl_issuers = []
            for crl_issuer in extension.crl_issuer:
                crl_issuers.append(_general_name_to_str(crl_issuer))

        return cls(
            full_name=full_names,
            name_relative_to_crl_issuer=relative_names,
            reasons=reasons,
            crl_issuer=crl_issuers,
        )

    def _string_dict(self):
        ret = {}
        if self.full_name is not None:
            ret["Full Name"] = self.full_name
        if self.name_relative_to_crl_issuer is not None:
            ret[
                "Name Relative To CRL Issuer"
            ] = self.name_relative_to_crl_issuer
        if self.reasons is not None:
            ret["Reasons"] = self.reasons
        if self.crl_issuer is not None:
            ret["CRL Issuer"] = self.crl_issuer
        return ret


class CrlDistributionPoints(Extension):
    crl_distribution_points: List[DistributionPoint]

    def __iter__(self) -> Iterable[DistributionPoint]:
        return iter(self.crl_distribution_points)

    @classmethod
    def from_cryptography(cls, extension: x509.CRLDistributionPoints):
        crl_distribution_points = []
        for crl_distribution_point in extension:
            parsed = DistributionPoint.from_cryptography(
                crl_distribution_point
            )
            crl_distribution_points.append(parsed)

        return cls(crl_distribution_points=crl_distribution_points)

    def _string_dict(self):
        ret = {self.name: []}
        for dist_point in self.crl_distribution_points:
            ret[self.name].append(dist_point._string_dict())

        return ret


class InhibitAnyPolicy(Extension):
    skip_certs: int

    @classmethod
    def from_cryptography(cls, extension: x509.InhibitAnyPolicy):
        return cls(skip_certs=extension.skip_certs)

    def _string_dict(self):
        return {self.name: {"Skip Certs": self.skip_certs}}


class FreshestCrl(CrlDistributionPoints):
    pass


class AccessDescription(Extension):
    access_method: str
    access_location: str

    @classmethod
    def from_cryptography(cls, extension: x509.AccessDescription):
        return cls(
            access_method=extension.access_method._name,
            access_location=_general_name_to_str(extension.access_location),
        )

    def _string_dict(self):
        return {
            "Access Method": self.access_method,
            "Access Location": self.access_location,
        }


class AuthorityInformationAccess(Extension):
    access_description: List[AccessDescription]

    def __iter__(self) -> Iterable[AccessDescription]:
        return iter(self.access_description)

    @classmethod
    def from_cryptography(cls, extension: x509.AuthorityInformationAccess):
        access_descriptions = []
        for access_description in extension:
            access_descriptions.append(
                AccessDescription.from_cryptography(access_description)
            )
        return cls(access_description=access_descriptions)

    def _string_dict(self):
        access_descriptions = []
        for access_description in self.access_description:
            access_descriptions.append(access_description._string_dict())

        return {self.name: {"Access Description": access_descriptions}}


class SubjectInformationAccess(AuthorityInformationAccess):
    pass


class Extensions(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    authority_key_identifier: Optional[AuthorityKeyIdentifier] = Field(
        alias=ExtensionOID.AUTHORITY_KEY_IDENTIFIER.dotted_string, default=None
    )
    subject_key_identifier: Optional[SubjectKeyIdentifier] = Field(
        alias=ExtensionOID.SUBJECT_KEY_IDENTIFIER.dotted_string, default=None
    )
    key_usage: Optional[KeyUsage] = Field(
        alias=ExtensionOID.KEY_USAGE.dotted_string, default=None
    )
    certificate_policies: Optional[CertificatePolicies] = Field(
        alias=ExtensionOID.CERTIFICATE_POLICIES.dotted_string, default=None
    )
    # policy_mappings
    subject_alternative_name: Optional[SubjectAlternativeName] = Field(
        alias=ExtensionOID.SUBJECT_ALTERNATIVE_NAME.dotted_string, default=None
    )
    issuer_alternative_name: Optional[IssuerAlternativeName] = Field(
        alias=ExtensionOID.ISSUER_ALTERNATIVE_NAME.dotted_string, default=None
    )
    subject_directory_attributes: Optional[SubjectDirectoryAttributes] = Field(
        alias=ExtensionOID.SUBJECT_DIRECTORY_ATTRIBUTES.dotted_string,
        default=None,
    )
    basic_constraints: Optional[BasicConstraints] = Field(
        alias=ExtensionOID.BASIC_CONSTRAINTS.dotted_string,
        default=None,
    )
    name_constraints: Optional[NameConstraints] = Field(
        alias=ExtensionOID.NAME_CONSTRAINTS.dotted_string,
        default=None,
    )
    policy_constraints: Optional[PolicyConstraints] = Field(
        alias=ExtensionOID.POLICY_CONSTRAINTS.dotted_string,
        default=None,
    )
    extended_key_usage: Optional[ExtendedKeyUsage] = Field(
        alias=ExtensionOID.EXTENDED_KEY_USAGE.dotted_string,
        default=None,
    )
    crl_distribution_points: Optional[CrlDistributionPoints] = Field(
        alias=ExtensionOID.CRL_DISTRIBUTION_POINTS.dotted_string,
        default=None,
    )
    inhibit_any_policy: Optional[InhibitAnyPolicy] = Field(
        alias=ExtensionOID.INHIBIT_ANY_POLICY.dotted_string,
        default=None,
    )
    freshest_crl: Optional[FreshestCrl] = Field(
        alias=ExtensionOID.FRESHEST_CRL.dotted_string,
        default=None,
    )
    authority_information_access: Optional[AuthorityInformationAccess] = Field(
        alias=ExtensionOID.AUTHORITY_INFORMATION_ACCESS.dotted_string,
        default=None,
    )
    subject_information_access: Optional[SubjectInformationAccess] = Field(
        alias=ExtensionOID.SUBJECT_INFORMATION_ACCESS.dotted_string,
        default=None,
    )

    @classmethod
    def from_cryptography(cls, cert_extensions: x509.Extensions):
        extensions_dict = {}

        for name, field_info in cls.model_fields.items():
            class_type = typing.get_args(field_info.annotation)[0]
            oid = ObjectIdentifier(field_info.alias)
            try:
                x509_ext = cert_extensions.get_extension_for_oid(oid)

                ext = class_type.from_cryptography(x509_ext.value)

                extensions_dict[oid.dotted_string] = ext
                extensions_dict[oid.dotted_string].critical = x509_ext.critical
            except ExtensionNotFound:
                logger.trace(f"Extension with OID: {oid._name} not found")

        return cls(**extensions_dict)

    @staticmethod
    def _get_extension_from_oid(
        cert_extensions: x509.Extensions, oid: ObjectIdentifier, classType
    ) -> Optional[ExtensionTypeVar]:
        try:
            ext_val = cert_extensions.get_extension_for_oid(oid).value
            classType.from_cryptography(ext_val)
        except ExtensionNotFound:
            logger.debug(f"Extension with OID: {oid._name} not found")
            return None

    def _string_dict(self):
        extensions = {}
        for field_name in self.model_fields:
            att_val = getattr(self, field_name)

            if att_val is None or str(att_val) == "":
                continue

            extensions.update(att_val._string_dict())

        return extensions


def _general_name_to_str(general_name):
    if general_name is None:
        return None
    name_str = f"{type(general_name).__name__}: "
    if isinstance(general_name, x509.OtherName):
        value = _byte_to_hex(general_name.value)
        name_str += f"{general_name.type_id.dotted_string} - {value}"
    elif isinstance(general_name, x509.RegisteredID):
        name_str += str(general_name.value.dotted_string)
    elif isinstance(general_name, x509.DirectoryName):
        name_str += str(Name.from_cryptography(general_name.value))
    else:
        name_str += str(general_name.value)
    return name_str


def _general_names_to_str(general_names: List[str]):
    names = ""
    for general_name in general_names:
        names += f"""
                     {general_name}"""
    return names



from .chain import Chain
from .certificates import Certificates
from .certificate import Certificate
from .crl import CertificateRevocationList
from .extensions import Extensions
from .name import Name



import time
from functools import lru_cache

from cryptography import x509
from loguru import logger


from pki_tools.utils import HTTPX_CLIENT, verify_signature

from pki_tools.types.chain import Chain

from pki_tools.types.certificate import Certificate

from pki_tools.exceptions import (
    ExtensionMissing,
    CrlFetchFailure,
    CrlLoadError,
)
from pki_tools.types.crl import CertificateRevocationList


def _is_revoked(
    cert: Certificate,
    crl_issuer: Chain,
    crl_cache_seconds: int = 3600,
) -> bool:
    crl_issuer.check_chain()
    logger.debug("CRL issuer chain valid")

    log = logger.bind(
        cert=cert.pem_string,
        serial=cert.hex_serial,
    )

    if cert.extensions.crl_distribution_points is None:
        log.debug("CRL extension missing")
        raise ExtensionMissing()

    http_dist = False
    for dist_point in cert.extensions.crl_distribution_points:
        if dist_point.full_name is None:
            continue

        for full_name in dist_point.full_name:
            split = full_name.split("UniformResourceIdentifier: ")
            if len(split) < 2:
                continue
            uri = split[1]

            http_dist = True
            cache_ttl = round(time.time() / crl_cache_seconds)
            crl = _get_crl_from_url(uri, cache_ttl=cache_ttl)

            issuer = crl_issuer.get_issuer(crl)

            verify_signature(crl, issuer)
            logger.debug("CRL signature valid")

            r = crl._x509_obj.get_revoked_certificate_by_serial_number(
                cert.serial_number,
            )
            if r is not None:
                log.bind(date=str(r.revocation_date)).debug(
                    "Certificate revoked"
                )
                return True

    if not http_dist:
        log.debug("CRL missing URI")
        raise ExtensionMissing()

    log.debug("Certificate valid")
    return False


@lru_cache(maxsize=None)
def _get_crl_from_url(crl_url, cache_ttl=None) -> CertificateRevocationList:
    ret = HTTPX_CLIENT.get(crl_url)

    if ret.status_code != 200:
        logger.bind(
            status=ret.status_code,
            url=crl_url,
        ).error("Failed to fetch CRL from URL")
        raise CrlFetchFailure()

    crl_data = ret.content
    return _crl_data_to_crypto(crl_data)


def _crl_data_to_crypto(crl_data) -> CertificateRevocationList:
    try:
        crypto_crl = x509.load_der_x509_crl(crl_data)
        return CertificateRevocationList.from_cryptography(crypto_crl)
    except (TypeError, ValueError):
        pass

    try:
        crypto_crl = x509.load_pem_x509_crl(crl_data)
        return CertificateRevocationList.from_cryptography(crypto_crl)
    except TypeError as e:
        logger.bind(crl=crl_data).error("Failed to load CRL")
        raise CrlLoadError(e) from None



import re
from typing import Union, Optional, Dict

import yaml

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, ec, dsa, rsa
from cryptography.hazmat.primitives.asymmetric.types import (
    CertificatePublicKeyTypes,
)

from pki_tools.types.name import Name
from pki_tools.types.extensions import Extensions

from pki_tools.exceptions import CertLoadError
from pki_tools.types.utils import _byte_to_hex

from typing import Type

from cryptography import x509

from pydantic import BaseModel


from pki_tools.types.crypto_parser import CryptoParser

from datetime import datetime

from loguru import logger
from pydantic import ConfigDict


PEM_REGEX = re.compile(
    r"\s*-+BEGIN CERTIFICATE-+[\w+/\s=]*-+END CERTIFICATE-+\s*"
)
CACHE_TIME_SECONDS = 60 * 60 * 24 * 30  # 1 month


class SignatureAlgorithm(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    algorithm: hashes.HashAlgorithm
    parameters: Union[None, padding.PSS, padding.PKCS1v15, ec.ECDSA] = None


class Validity(BaseModel):
    not_before: datetime
    not_after: datetime

    def _string_dict(self):
        return {
            "Not Before": self.not_before,
            "Not After": self.not_after,
        }


class SubjectPublicKeyInfo(CryptoParser):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    algorithm: str
    parameters: Dict[str, str]

    @classmethod
    def from_cryptography(cls, cert_public_key: CertificatePublicKeyTypes):
        name = str(cert_public_key.__class__).split(".")[-2].upper()
        parameters = {}
        if isinstance(cert_public_key, dsa.DSAPublicKey):
            pub_numbers = cert_public_key.public_numbers()
            param_numbers = pub_numbers.parameter_numbers
            parameters = {
                "key_size": str(cert_public_key.key_size),
                "public_key_y": pub_numbers.y,
                "prime_p": param_numbers.p,
                "subprime_q": param_numbers.q,
                "generator_g": param_numbers.g,
            }
        elif isinstance(cert_public_key, rsa.RSAPublicKey):
            pub_numbers = cert_public_key.public_numbers()
            parameters = {
                "key_size": str(cert_public_key.key_size),
                "modulus_n": str(pub_numbers.n),
                "exponent_e": str(pub_numbers.e),
            }
        elif isinstance(cert_public_key, ec.EllipticCurvePublicKey):
            pub_numbers = cert_public_key.public_numbers()
            parameters = {
                "key_size": str(cert_public_key.key_size),
                "x_coordinate": str(pub_numbers.x),
                "y_coordinate": str(pub_numbers.y),
                "curve": pub_numbers.curve.name,
            }

        return cls(algorithm=name, parameters=parameters)

    def _string_dict(self):
        params = {}
        for k, v in self.parameters.items():
            key = " ".join(ele.title() for ele in k.split("_"))
            params[key] = v

        return {"Public Key Algorithm": self.algorithm, "Parameters": params}


class TbsCertificate(BaseModel):
    version: int
    serial_number: int
    signature_algorithm: SignatureAlgorithm
    issuer: Name
    validity: Validity
    subject: Name
    subject_public_key_info: SubjectPublicKeyInfo
    extensions: Optional[Extensions]

    def _string_dict(self):
        return {
            "Version": self.version,
            "Serial Number": self.hex_serial,
            "Signature Algorithm": self.signature_algorithm.algorithm.name,
            "Issuer": str(self.issuer),
            "Validity": self.validity._string_dict(),
            "Subject": str(self.subject),
            "Subject Public Key Info": self.subject_public_key_info._string_dict(),
            "Extensions": self.extensions._string_dict(),
        }

    @property
    def hex_serial(self) -> str:
        """
        Parses the certificate serial into hex format

        Returns:
            String representing the hex value of the certificate serial number
        """
        hex_serial = format(self.serial_number, "x").zfill(32)
        return hex_serial.upper()

    @property
    def public_key(self) -> bytes:
        return self.subject_public_key_info.public_bytes(
            serialization.Encoding.Raw,
            serialization.PublicFormat.Raw,
        )


class Certificate(TbsCertificate, CryptoParser):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    signature_value: str

    @classmethod
    def from_cryptography(
        cls: Type["Certificate"], cert: x509.Certificate
    ) -> "Certificate":
        ret = cls(
            version=cert.version.value,
            serial_number=cert.serial_number,
            signature_algorithm=SignatureAlgorithm(
                algorithm=cert.signature_hash_algorithm,
                parameters=cert.signature_algorithm_parameters,
            ),
            issuer=Name.from_cryptography(cert.issuer),
            validity=Validity(
                not_before=cert.not_valid_before,
                not_after=cert.not_valid_after,
            ),
            subject=Name.from_cryptography(cert.subject),
            subject_public_key_info=SubjectPublicKeyInfo.from_cryptography(
                cert.public_key()
            ),
            extensions=Extensions.from_cryptography(cert.extensions),
            signature_value=_byte_to_hex(cert.signature),
        )
        ret._x509_obj = cert
        return ret

    @classmethod
    def from_pem_string(cls: Type["Certificate"], cert_pem) -> "Certificate":
        """
        Loads a certificate from a PEM string into a
        [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
        object

        Arguments:
            cert_pem -- The PEM encoded certificate in string format
        Returns:
            A
            [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
            created from the PEM
        Raises:
             exceptions.CertLoadError - If the certificate could not be loaded
        """
        try:
            cert_pem = re.sub(r"\n\s*", "\n", cert_pem)
            if not _is_pem_string(cert_pem):
                raise ValueError

            crypto_cert = x509.load_pem_x509_certificate(cert_pem.encode())
            return Certificate.from_cryptography(crypto_cert)
        except ValueError as e:
            logger.bind(cert=cert_pem).debug("Failed to load cert from PEM")
            raise CertLoadError(e)

    @classmethod
    def from_file(cls: Type["Certificate"], file_path: str) -> "Certificate":
        """
        Reads a file containing one PEM certificate into a
        [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
        object

        Arguments:
            file_path -- Path and filename of the PEM certificate
        Returns:
             The
             [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
             representing the certificate from file
        """

        with open(file_path, "r") as f:
            cert_pem = f.read()

        return Certificate.from_pem_string(cert_pem)

    @property
    def tbs_bytes(self) -> bytes:
        return self._x509_obj.tbs_certificate_bytes

    def to_file(self, file_path):
        with open(file_path, "w") as f:
            f.write(self.pem_string)

    def _string_dict(self):
        return {
            "Certificate": {
                "TbsCertificate": super()._string_dict(),
                "Signature Value": self.signature_value,
            }
        }

    @property
    def pem_string(self):
        return self._x509_obj.public_bytes(serialization.Encoding.PEM).decode()

    @property
    def public_key(self) -> CertificatePublicKeyTypes:
        return self._x509_obj.public_key()

    def __str__(self) -> str:
        return yaml.safe_dump(
            self._string_dict(),
            indent=2,
            default_flow_style=False,
            explicit_start=True,
            default_style="",
        )


def _is_pem_string(check: str):
    if not isinstance(check, str):
        return False

    return re.match(PEM_REGEX, check)



from typing import Type

from cryptography import x509

from pki_tools.types.crypto_parser import CryptoParser
from pki_tools.types.name import Name


class CertificateRevocationList(CryptoParser):
    issuer: Name

    @classmethod
    def from_cryptography(
        cls: Type["CertificateRevocationList"],
        crypto_crl: x509.CertificateRevocationList,
    ) -> "CertificateRevocationList":
        ret = cls(issuer=Name.from_cryptography(crypto_crl.issuer))
        ret._x509_obj = crypto_crl
        return ret

    @property
    def tbs_bytes(self) -> bytes:
        return self._x509_obj.tbs_certlist_bytes



import httpx


from cryptography.hazmat.primitives.asymmetric import padding

from cryptography.x509.ocsp import OCSPResponse
from loguru import logger

from pki_tools.types.certificate import Certificate
from pki_tools.exceptions import SignatureVerificationFailed
from pki_tools.types.crl import CertificateRevocationList

HTTPX_CLIENT = httpx.Client(
    transport=httpx.HTTPTransport(retries=2), timeout=15
)


def verify_signature(
    signed: [Certificate, CertificateRevocationList, OCSPResponse],
    issuer: Certificate,
) -> None:
    """
    Verifies a signature of a signed entity against the issuer certificate

    Args:
        signed: The signed certificate can either be a
        [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
        [CertificateRevocationList](https://pki-tools.fulder.dev/pki_tools/types/#certificaterevocationlist)
        or a
        [OCSPResponse](https://pki-tools.fulder.dev/pki_tools/types/#ocspresponse)
        issuer: The
        [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
        describing the issuer of the signed entity
    Raises:
        [InvalidSignedType](https://pki-tools.fulder.dev/pki_tools/exceptions/#invalidsignedtype)
        -- When the issuer has a non-supported type
        [SignatureVerificationFailed](https://pki-tools.fulder.dev/pki_tools/exceptions/#signatureverificationfailed)
        -- When the signature verification fails
    """
    try:
        issuer.public_key.verify(
            signed._x509_obj.signature,
            signed.tbs_bytes,
            padding.PKCS1v15(),
            signed._x509_obj.signature_hash_algorithm,
        )
        logger.trace("Signature valid")
    except Exception as e:
        logger.bind(
            exceptionType=type(e).__name__,
            exception=str(e),
        ).error("Signature verification failed")
        raise SignatureVerificationFailed(
            f"signature doesn't match issuer "
            f"with subject: {str(issuer.subject)}"
        )



from loguru import logger

from .types import Chain, Certificate, Certificates, Extensions, Name

from .exceptions import (
    ExtensionMissing,
    Error,
    OcspInvalidResponseStatus,
    OcspFetchFailure,
    OcspIssuerFetchFailure,
    RevokeCheckFailed,
    NotCompleteChain,
    CertIssuerMissingInChain,
    CertExpired,
    InvalidSignedType,
    SignatureVerificationFailed,
    CertLoadError,
)

from .crl import _is_revoked
from .ocsp import _is_revoked_multiple_issuers


def is_revoked(
    cert: Certificate,
    chain: Chain,
    crl_cache_seconds: int = 3600,
) -> bool:
    """
    Checks if a certificate is revoked first using the OCSP extension and then
    the CRL extensions.

    Note that OCSP has precedence over CRL meaning that if OCSP check is
    successful this function will return the bool without checking CRL.

    Otherwise, if OCSP check fails, CRL will be tried next.

    Arguments:
        cert -- The
        [Certificates](https://pki-tools.fulder.dev/pki_tools/types/#certificates)
        certificate to check revocation for.
        chain -- The CA chain including one or more certificates and
        the issuer of the `cert`, signer of the OCSP response and CRL
        issuer. See
        [types.Chain](https://pki-tools.fulder.dev/pki_tools/types/#chain)
        for examples how the chain can be created
        crl_cache_seconds -- [CRL Only] Specifies how long the CRL should
        be cached, default is 1 hour.
    Returns:
        True if the certificate is revoked, False otherwise
    Raises:
        [exceptions.ChainVerificationFailed](https://pki-tools.fulder.dev/pki_tools/exceptions/#chainverificationfailed)
        -- When the Chain contains more than one certificate and
        the trust fails either because of some certificate has expired
        or some signature in the chain is invalid

        [exceptions.RevokeCheckFailed](https://pki-tools.fulder.dev/pki_tools/exceptions/#revokecheckfailed)
        -- When both OCSP and CRL checks fail
    """
    return is_revoked_multiple_issuers(
        cert, chain, chain, chain, crl_cache_seconds
    )


def is_revoked_multiple_issuers(
    cert: Certificate,
    cert_issuer: Chain,
    ocsp_issuer: Chain,
    crl_issuer: Chain,
    crl_cache_seconds: int = 3600,
):
    """
    Checks if a certificate is revoked first using the OCSP extension and then
    the CRL extensions.

    Note that OCSP has precedence over CRL meaning that if OCSP check is
    successful this function will return the bool without checking CRL.

    Otherwise, if OCSP check fails, CRL will be tried next.

    Arguments:
        cert -- The certificate to check revocation for. Can either be
        a
        [Certificate](https://cryptography.io/en/latest/x509/reference/#cryptography.x509.Certificate)
        or a
        [types.PemCert](https://pki-tools.fulder.dev/pki_tools/types/#pemcert)
        string
        cert_issuer -- The CA chain including one or more certificates and
        the issuer of the `cert`. See
        [types.Chain](https://pki-tools.fulder.dev/pki_tools/types/#chain)
        for examples how the chain can be created
        ocsp_issuer -- The CA chain including one or more certificates used
        for signing of the OCSP response
        crl_issuer -- The CA chain including one or more certificates used
        for signing the CRL
        crl_cache_seconds -- [CRL Only] Specifies how long the CRL should be
        cached, default is 1 hour.
    Returns:
        True if the certificate is revoked, False otherwise
    Raises:
        [exceptions.ChainVerificationFailed](https://pki-tools.fulder.dev/pki_tools/exceptions/#chainverificationfailed)
        -- When the Chain contains more than one certificate and
        the trust fails either because of some certificate has expired
        or some signature in the chain is invalid

        [exceptions.RevokeCheckFailed](https://pki-tools.fulder.dev/pki_tools/exceptions/#revokecheckfailed)
        -- When both OCSP and CRL checks fail
    """

    try:
        return _is_revoked_multiple_issuers(cert, cert_issuer, ocsp_issuer)
    except (
        ExtensionMissing,
        OcspInvalidResponseStatus,
        OcspFetchFailure,
        OcspIssuerFetchFailure,
    ):
        logger.debug("OCSP revoke check failed, trying CRL next")

    try:
        return _is_revoked(cert, crl_issuer)
    except Error as e:
        err_message = "OCSP and CRL checks failed"
        logger.bind(exceptionType=type(e).__name__).error(err_message)
        raise RevokeCheckFailed(err_message) from None



import hashlib
from typing import Type

from cryptography.x509 import ocsp
from loguru import logger

from pki_tools.types.crypto_parser import CryptoParser
from pki_tools.types.utils import _byte_to_hex


class OCSPResponse(CryptoParser):
    response_status: str
    certificate_status: str
    issuer_key_hash: str

    @classmethod
    def from_cryptography(
        cls: Type["OCSPResponse"], crypto_ocsp_response: ocsp.OCSPResponse
    ) -> "OCSPResponse":
        try:
            ocsp_response_key_hash = _byte_to_hex(
                crypto_ocsp_response.issuer_key_hash
            )
        except Exception as e:
            logger.bind(
                exceptionType=type(e),
                exception=str(e),
                issuerHash=crypto_ocsp_response.issuer_key_hash,
            ).error("Couldn't convert issuer key hash to hex")
            raise

        ret = cls(
            response_status=crypto_ocsp_response.response_status.name,
            certificate_status=crypto_ocsp_response.certificate_status.name,
            issuer_key_hash=ocsp_response_key_hash,
        )
        ret._x509_obj = crypto_ocsp_response
        return ret

    @property
    def tbs_bytes(self) -> bytes:
        return self._x509_obj.tbs_response_bytes

    def hash_with_alg(self, der_key) -> str:
        hash_algorithm = hashlib.new(self._x509_obj.hash_algorithm.name)
        hash_algorithm.update(der_key)
        return hash_algorithm.hexdigest().upper()

    @property
    def is_successful(self):
        return self.response_status == "SUCCESSFUL"

    @property
    def is_revoked(self):
        return self.certificate_status == "REVOKED"



import time
from functools import lru_cache
from typing import List, Type, T

import httpx
from cryptography import x509
from loguru import logger

from pydantic import BaseModel, constr


from pki_tools.exceptions import OcspIssuerFetchFailure
from pki_tools.types.certificate import Certificate
from pki_tools.types.crypto_parser import CryptoParser

CACHE_TIME_SECONDS = 60 * 60 * 24 * 30  # 1 month


class CertsUri(BaseModel):
    """
    Describes a URI where one or more public certificate(s)
    can be downloaded

    Examples::
        CertsUri(uri="https://my.ca.link.com/ca.pem")
    Attributes:
        uri -- The URI for the public certificate(s)
        cache_time_seconds -- Specifies how long the public cert should be
        cached, default is 1 month.
    """

    uri: constr(pattern=r"https*://.*")
    cache_time_seconds: int = CACHE_TIME_SECONDS


class Certificates(CryptoParser):
    """
    Attributes:
        certificates -- a list of
        [Certificates](https://pki-tools.fulder.dev/pki_tools/types/#certificates)
    """

    certificates: List[Certificate]

    @classmethod
    def from_cryptography(
        cls: Type[T], crypto_certs: List[x509.Certificate]
    ) -> T:
        certificates = []
        for crypt_cert in crypto_certs:
            certificates.append(Certificate.from_cryptography(crypt_cert))
        return cls(certificates=certificates)

    @classmethod
    def from_file(cls: T, file_path: str) -> T:
        """
        Reads a file containing one or more PEM certificate(s) into a
        [Certificates](https://pki-tools.fulder.dev/pki_tools/types/#certificates)
        object

        Arguments:
            file_path -- Path and filename of the PEM certificate
        Returns:
             A
             [Certificates](https://pki-tools.fulder.dev/pki_tools/types/#certificates)
             object representing the certificate(s) from file
        """
        with open(file_path, "r") as f:
            cert_pem = f.read()

        crypto_certs = x509.load_pem_x509_certificates(cert_pem.encode())
        return cls.from_cryptography(crypto_certs)

    @classmethod
    def from_pem_string(cls: T, pem_string: str) -> T:
        crypto_certs = x509.load_pem_x509_certificates(pem_string.encode())
        return cls.from_cryptography(crypto_certs)

    @classmethod
    def from_uri(
        cls: T,
        uri: str,
        cache_time_seconds: int = CACHE_TIME_SECONDS,
    ) -> T:
        """
        Loads
        [Certificates](https://pki-tools.fulder.dev/pki_tools/types/#certificates)
        from a str URI

        Arguments:
             uri -- A str containing the URI where the certificate(s)
             can be downloaded.
             cache_time_seconds -- Decides how long the certificates
             should be cached, default is 1 month
        """
        chain_uri = CertsUri(uri=uri, cache_time_seconds=cache_time_seconds)
        cache_ttl = round(time.time() / chain_uri.cache_time_seconds)
        return cls._from_uri(chain_uri.uri, cache_ttl)

    @classmethod
    @lru_cache(maxsize=None)
    def _from_uri(cls: T, uri: str, ttl=None) -> T:
        ret = httpx.get(uri)

        if ret.status_code != 200:
            logger.bind(status=ret.status_code).error(
                "Failed to fetch issuer from URI"
            )
            raise OcspIssuerFetchFailure(
                f"Issuer URI fetch failed. Status: {ret.status_code}"
            )

        return cls(certificates=x509.load_pem_x509_certificate(ret.content))

    @property
    def pem_string(self):
        all_certs = ""
        for cert in self.certificates:
            all_certs += cert.pem_string

        return all_certs

    def to_file(self, file_path: str):
        """
        Saves one or more certificate(s) into a file

        Arguments:
            file_path -- Path and filename where to store the certificate(s)
        """
        with open(file_path, "w") as f:
            f.write(self.pem_string)

        logger.debug(f"Certificate(s) saved to {file_path}")



from collections import defaultdict
from typing import List

from cryptography import x509
from cryptography.hazmat._oid import NameOID


from pydantic import Field, ConfigDict

from pki_tools.types.crypto_parser import CryptoParser


class Name(CryptoParser):
    """
    Name type describes certificate subject or issuer.
    The attributes are following the
    [RFC5280#Section-4.1.2.4](https://datatracker.ietf.org/doc/html/rfc5280#section-4.1.2.4)

    Note that every attribute is a list of string in order to support
    multivalued RDNs.

    Attributes:
        c -- Country Name (2.5.4.6)
        o -- Organization Name (2.5.4.10)
        ou -- Organizational Unit Name (2.5.4.11)
        dnq -- Distinguished Name Qualifier (2.5.4.46)
        s -- State Or Province Name (2.5.4.8)
        cn -- Common Name (2.5.4.3)
        serial -- Serial Number (2.5.4.5)
        ln -- Locality Name (2.5.4.7)
        t -- Title (2.5.4.12)
        sn -- Surname (2.5.4.4)
        gn -- Given Name (2.5.4.42)
        i -- Initials (2.5.4.43)
        p -- Pseudonym (2.5.4.65)
        gq -- Generation Qualifier (2.5.4.44)
        dc -- Domain Component (0.9.2342.19200300.100.1.25)
    """

    model_config = ConfigDict(populate_by_name=True)

    c: List[str] = Field(alias=NameOID.COUNTRY_NAME.dotted_string, default=[])
    o: List[str] = Field(
        alias=NameOID.ORGANIZATION_NAME.dotted_string, default=[]
    )
    ou: List[str] = Field(
        alias=NameOID.ORGANIZATIONAL_UNIT_NAME.dotted_string, default=[]
    )
    dnq: List[str] = Field(
        alias=NameOID.DN_QUALIFIER.dotted_string, default=[]
    )
    s: List[str] = Field(
        alias=NameOID.STATE_OR_PROVINCE_NAME.dotted_string, default=[]
    )
    cn: List[str] = Field(alias=NameOID.COMMON_NAME.dotted_string, default=[])
    serial: List[str] = Field(
        alias=NameOID.SERIAL_NUMBER.dotted_string, default=[]
    )

    ln: List[str] = Field(
        alias=NameOID.LOCALITY_NAME.dotted_string, default=[]
    )
    t: List[str] = Field(alias=NameOID.TITLE.dotted_string, default=[])
    sn: List[str] = Field(alias=NameOID.SURNAME.dotted_string, default=[])
    gn: List[str] = Field(alias=NameOID.GIVEN_NAME.dotted_string, default=[])
    i: List[str] = Field(alias=NameOID.INITIALS.dotted_string, default=[])
    p: List[str] = Field(alias=NameOID.PSEUDONYM.dotted_string, default=[])
    gq: List[str] = Field(
        alias=NameOID.GENERATION_QUALIFIER.dotted_string, default=[]
    )
    dc: List[str] = Field(
        alias=NameOID.DOMAIN_COMPONENT.dotted_string, default=[]
    )

    @classmethod
    def from_cryptography(cls, name: x509.Name):
        subject = defaultdict(set)
        for attribute in name:
            for att in name.get_attributes_for_oid(attribute.oid):
                subject[att.oid.dotted_string].add(att.value)
        return cls(**subject)

    def to_crypto_name(self) -> x509.Name:
        name_list = []
        for attr_name in vars(self):
            vals = getattr(self, attr_name)
            if not vals:
                continue

            oid = Name.model_fields[attr_name].alias
            for val in vals:
                name_list.append(
                    x509.NameAttribute(x509.ObjectIdentifier(oid), val)
                )

        return x509.Name(name_list)

    def _string_dict(self):
        ret = {}
        for a in self.model_dump():
            for val in getattr(self, a):
                ret[a.upper()] = val
        return ret

    def __str__(self):
        name_list = []
        for k, v in self._string_dict().items():
            name_list.append(f"{k}: {v}")
        return ", ".join(name_list)



class Error(Exception):
    pass


class ExtensionMissing(Error):
    pass


class CertLoadError(Error):
    pass


class OcspFetchFailure(Error):
    pass


class OcspInvalidResponseStatus(Error):
    pass


class OcspIssuerFetchFailure(Error):
    pass


class CrlFetchFailure(Error):
    pass


class CrlLoadError(Error):
    pass


class RevokeCheckFailed(Error):
    pass


class CertIssuerMissingInChain(Error):
    pass


class NotCompleteChain(Error):
    pass


class CertExpired(Error):
    pass


class SignatureVerificationFailed(Error):
    pass


class InvalidSignedType(Error):
    pass



from datetime import datetime

from loguru import logger
from pydantic import ConfigDict

from pki_tools.exceptions import (
    NotCompleteChain,
    CertExpired,
    CertIssuerMissingInChain,
)
from pki_tools.types.certificate import Certificate
from pki_tools.types.certificates import Certificates
from pki_tools.types.crl import CertificateRevocationList
from pki_tools.utils import verify_signature


class Chain(Certificates):
    """
    Chain holds a list of certificates in a
    [chain of trust](https://en.wikipedia.org/wiki/Chain_of_trust)

    Examples:
    From File::
        chain = Chain.from_file("/path/to/chain.pem")
    From PEM::
        pem_string="-----BEGIN CERTIFICATE-----...."
        chain = Chain.from_pem_string(pem_string)
    From URI::
        chain = Chain.from_uri("https://chain.domain/chain.pem")
    Using Chain::
        cert: Certificate = ...
        chain.check_chain()
        chain.get_issuer(cert)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def check_chain(self):
        """
        Validate the chain (if it contains more than one certificate)
        checking expiration and signatures of all certificates in the chain

        Raises:
            [exceptions.NotCompleteChain](https://pki-tools.fulder.dev/pki_tools/exceptions/#notcompletechain)
            -- When the chain contain only one not self-signed certificate

            [exceptions.CertExpired](https://pki-tools.fulder.dev/pki_tools/exceptions/#certexpired)
            -- If some certificate in the chain has expired

            [exceptions.InvalidSignedType](https://pki-tools.fulder.dev/pki_tools/exceptions/#invalidsignedtype)
            -- When the issuer has a non-supported type

            [exceptions.SignatureVerificationFailed](https://pki-tools.fulder.dev/pki_tools/exceptions/#signatureverificationfailed)
            -- When the signature verification fails
        """
        if len(self.certificates) == 1:
            if self.certificates[0].issuer == self.certificates[0].subject:
                logger.debug(
                    "Chain contains only one self signed cert, "
                    "nothing to check"
                )
                return
            else:
                raise NotCompleteChain()

        for cert in self.certificates:
            log = logger.bind(subject=cert.subject)
            if (
                cert.not_valid_after < datetime.now()
                or cert.not_valid_before > datetime.now()
            ):
                log.error("Certificate expired")
                raise CertExpired(
                    f"Certificate in chain with subject: '{cert.subject}' "
                    f"has expired"
                )

            issuer = self.get_issuer(cert)

            verify_signature(cert, issuer)

    def get_issuer(
        self,
        signed: [
            Certificate,
            CertificateRevocationList,
        ],
    ) -> Certificate:
        """
        Returns the issuer of a signed entity

        Arguments:
            signed: The signed entity can either be a
            [Certificate](https://pki-tools.fulder.dev/pki_tools/types/#certificate)
            [CertificateRevocationList](https://pki-tools.fulder.dev/pki_tools/types/#certificaterevocationlist)
        Returns:
            The
            [Certificate](https://cryptography.io/en/latest/x509/reference/#cryptography.x509.Certificate)
            representing the issuer of the `signed` entity
        Raises:
            [CertIssuerMissingInChain](https://pki-tools.fulder.dev/pki_tools/exceptions/#certissuermissinginchain)
            -- When the issuer of the entity is missing in the chain
        """
        cert_subject = signed.issuer
        log = logger.bind(subject=cert_subject._string_dict())

        for next_chain_cert in self.certificates:
            if cert_subject == next_chain_cert.subject:
                log.trace("Found issuer cert in chain")
                return next_chain_cert

        raise CertIssuerMissingInChain()
